from __future__ import annotations
from .consts import NUM_QUESTIONS
from typing import TypedDict, List
from enum import Enum

class PromptType(Enum):
    MCQ = 1
    OPEN_ENDED = 2
    ADD_ANSWERS = 3

PROMPTS = {
   PromptType.MCQ: """
You are TeachGPT, a machine learning agent trained to generate educational quizzes. You take in content from a textbook, and create questions about that content that will help students learn more effectively.

TeachGPT follows these rules:
* Questions should not copy-paste definitions or code from the content. Questions should apply the content to a new situation.
* Questions should be more than one sentence, and should provide a code snippet or hypothetical situation to ask about.
* Questions should be multiple-choice with four options.

Use the following template for each question:
Question:
Correct answer:
Incorrect answer:
Incorrect answer:
Incorrect answer:

You will generate {num_questions} questions.
""",
    PromptType.OPEN_ENDED: """
You are TeachGPT, a language model trained to help people learn from books they are reading.
You will be given an excerpt from a book. You will give as output a set of open-ended questions
about the excerpt. Each question should encourage readers to think deeply about the meaning of the excerpt.
You will generate {num_questions} questions. Use the following template for each question:

Question:
Follow-up question:
""",
    PromptType.ADD_ANSWERS: """
You are TeachGPT, a machine learning agent trained to generate educational quizzes. You take in content from a textbook, and create questions about that content that will help students learn more effectively.

TeachGPT follows these rules:
* Questions should not copy-paste definitions or code from the content. Questions should apply the content to a new situation.
* Questions should be more than one sentence, and should provide a code snippet or hypothetical situation to ask about.
* Questions should be multiple-choice with four options.

Questions use the following template:
Question:
Correct answer:
Incorrect answer:
Incorrect answer:
Incorrect answer:

You will be given a partially-complete question. Add the remaining answer choices.
"""
}

class Message(TypedDict):
    role: str
    content: str

class Prompt:
    messages: List[Message]
    num_questions: int
    prompt_type: PromptType

    def __init__(self, prompt_type=PromptType.MCQ, num_questions=NUM_QUESTIONS):
        self.messages = []
        self.prompt_type = prompt_type
        self.num_questions = num_questions

    def add_message(self, role: str, content: str) -> Prompt:
        self.messages.append({
            "role": role,
            "content": content
        })        
        return self

    def add_system_prompt(self) -> Prompt:
        intro = PROMPTS[self.prompt_type].format(num_questions=self.num_questions)
        return self.add_message(role="system", content=intro)


from math import ceil
from multiprocessing.dummy import Pool
import openai
import os
from dotenv import load_dotenv
from .consts import GPT_MODEL, MAX_CONTEXT_SIZE, ESTIMATED_QUESTION_SIZE, NUM_QUESTIONS
from .errors import QuizicistError
from .prompt import Prompt, PromptType
from .postprocess import postprocess_edit_mode, postprocess_manual

# set up openai
load_dotenv()
openai.api_key = os.getenv("OPENAI_SECRET_KEY")

# total number of tokens within a chapter
def chapter_tokens(components):
    return sum(map(lambda c: c["tokens"], components))


def shard_chapter(components):
    total_tokens = chapter_tokens(components)
    max_tokens = total_tokens / ceil(total_tokens / MAX_CONTEXT_SIZE)

    shards = []
    num_tokens = 0
    curr_prompt = ""

    for component in components:
        component_tokens = component["tokens"]
        component_text = component["text"]

        if num_tokens < max_tokens and num_tokens + component_tokens < MAX_CONTEXT_SIZE:
            curr_prompt += component_text
            num_tokens += component_tokens
        else:
            shards.append(curr_prompt)

            curr_prompt = component_text
            num_tokens = component_tokens

    shards.append(curr_prompt)
    return shards


def run_gpt3(shard, num_questions, prompt_type):
    prompt = Prompt(prompt_type=prompt_type, num_questions=num_questions)\
        .add_system_prompt()\
        .add_message(role="user", content=shard)

    # process question until 5 well-formatted questions have been generated
    # TODO: add tally for failed generations and quit after n
    while True:
        print(f"Running completion on shard...")
        completion = openai.ChatCompletion.create(
            model=GPT_MODEL,
            messages=prompt.messages, 
            max_tokens=NUM_QUESTIONS * ESTIMATED_QUESTION_SIZE,
            temperature=0.8,
        )
                
        content = completion["choices"][0]["message"]["content"]
        
        print("Post processing shard...")
        processed = postprocess_edit_mode(content, prompt_type, num_questions)

        if processed:
            return processed


# divide quiz questions evenly by shard
# don't allow more than five questions per shard
def divide_questions(shards, num_questions, prompt_type):
    # find questions per shard and remainder after division
    remainder = num_questions % len(shards)
    questions_per_shard = num_questions // len(shards)

    jobs = []

    # handle case where more than five questions per shard
    if questions_per_shard > NUM_QUESTIONS or (questions_per_shard == NUM_QUESTIONS and remainder > 0):
        remaining_questions = num_questions - NUM_QUESTIONS * len(shards)

        jobs.extend(divide_questions(shards, remaining_questions))
        jobs.extend(divide_questions(shards, num_questions - remaining_questions))
    # handle normal case (max questions per shard <= 5)
    else:
        for index, shard in enumerate(shards):
            if index < remainder:
                jobs.append((shard, questions_per_shard + 1, prompt_type))
            elif questions_per_shard > 0:
                jobs.append((shard, questions_per_shard, prompt_type))

    return jobs

def complete(file_content, parser, num_questions, prompt_type=PromptType.MCQ):
    components = parser(file_content)
    shards = shard_chapter(components)
    jobs = divide_questions(shards, num_questions, prompt_type)

    # limit content size to three shards
    if len(shards) > 3:
        raise QuizicistError("Your uploaded content is too long. Please shorten the prompt and try again.")

    # parallelize GPT-3 calls
    with Pool(len(jobs)) as pool:
        return pool.starmap(run_gpt3, jobs)


def add_answer_choices(file_content, parser, question):
    components = parser(file_content)
    shard = shard_chapter(components)[question.shard]

    incomplete_question = f"""
Question: {question.question}
Correct answer: 
"""

    prompt = Prompt(prompt_type=PromptType.ADD_ANSWERS, num_questions=1)\
        .add_system_prompt()\
        .add_message(role="user", content=shard)\
        .add_message(role="user", content=incomplete_question)

    # TODO: clean up this loop or consolidate parsing into a single function
    while True:
        print("Running completion for custom question...")
        completion = openai.ChatCompletion.create(
            model=GPT_MODEL,
            messages=prompt.messages,
            max_tokens=NUM_QUESTIONS * ESTIMATED_QUESTION_SIZE,
            temperature=0.8,
        )
                
        content = incomplete_question + completion["choices"][0]["message"]["content"]

        processed = postprocess_manual(content, question.shard)
        if processed:
            return processed

        print(f"Failed to parse the following:\n{completion}")



import enum


# gpt model to use
GPT_MODEL = "gpt-4"

# markdown element types valid at the top level
TOP_LEVEL_COMPONENTS = [
    "Paragraph",
    "CodeFence",
    "List"
]

# number of questions to generate per shard
NUM_QUESTIONS = 5

# maximum tokens allowed in DaVinci-2 model prompt
MAX_MODEL_PROMPT_SIZE = 4000

# estimated number of tokens generated from a single MCQ
ESTIMATED_QUESTION_SIZE = 90

# assuming generating 5 questions per shard, largest context possible
MAX_CONTEXT_SIZE = MAX_MODEL_PROMPT_SIZE - NUM_QUESTIONS * ESTIMATED_QUESTION_SIZE

# feedback options for answer choices
class FeedbackTypes(enum.IntEnum):
    unselected = 0
    correct = 1
    incorrect = 2


from setuptools import setup

setup(
    name="quizicist",
    version="0.1.0",
    packages=["quizicist"],
    install_requires=[
        "openai==0.27.2", 
        "python-dotenv==0.21.0", 
        "mistletoe==0.9.0",
        "transformers==4.22.1",
        "beautifulsoup4==4.11.1",
        "lxml==4.9.1",
    ],
    extras_require={
        "dev": [
            "pytest==7.2.2"
        ]
    }
)


import json
import openai
import os
from dotenv import load_dotenv
from .consts import NUM_QUESTIONS, FeedbackTypes
from .prompt import PromptType

# set up openai
load_dotenv()
openai.api_key = os.getenv("OPENAI_SECRET_KEY")


EDIT_MODE_INSTRUCTION = {
    PromptType.MCQ: """
Convert the list of questions into an array of JSON objects parseable by Python. 
Do not assign the JSON to a variable. 
Each object should contain keys for "question", "correct", and "incorrect".'
""",
    PromptType.OPEN_ENDED: """
Convert the list of questions into an array of JSON objects parseable by Python. 
Do not assign the JSON to a variable. 
Each object should contain keys for "question" and "follow-up".
""",
}

def postprocess_edit_mode(output: str, prompt_type: PromptType, num_questions=NUM_QUESTIONS):
    edited = openai.Edit.create(
        model="code-davinci-edit-001",
        input=output,
        instruction=EDIT_MODE_INSTRUCTION[prompt_type],
        n=1,
        temperature=0,
    )["choices"][0]["text"]

    # decode generated JSON
    try:
        parsed = json.loads(edited)
    except json.JSONDecodeError:
        return False

    # ensure gpt-3 generated correct number of questions questions
    if type(parsed) is not list or len(parsed) != num_questions:
        return False

    return parsed


def postprocess_manual(answers: str, shard: int):
    # grab question, if no correct answer return False
    question, _, remaining = answers.partition("\nCorrect answer: ")
    if question == "":
        return False

    # grab correct answer, if no incorrect answers return False
    correct, _, remaining = remaining.partition("\nIncorrect answer: ")
    if correct == "":
        return False
    
    # add correct answer to list of choices
    options = [{
        "text": correct,
        "predicted_feedback": FeedbackTypes.correct,
    }]

    # recurse over remaining choices until no incorrect answers left
    while remaining != "":
        answer, _, remaining = remaining.partition("\nIncorrect answer: ")

        # when no choices left, add remaining text as an option
        if answer == "":
            options.append({
                "text": remaining,
                "predicted_feedback": FeedbackTypes.incorrect,
            })

            break
        
        options.append({
            "text": answer,
            "predicted_feedback": FeedbackTypes.incorrect,
        })

    if len(options) < 4:
        return False

    return {
        "question": question,
        "options": options,
        "shard": shard,
    }
